#!/usr/bin/env python3
# system modules
import logging
import time
from math import floor

# internal modules
from patatmo import utils
from patatmo.api import authentication
from patatmo.api import responsetypes
from patatmo.api import requests
from patatmo.api.variables import *
from patatmo.api.errors import *

# external modules
import numpy as np


class NetatmoClient(object):
    """
    Netatmo api client

    Args:
        authentication (Authentication, optional):
            the authentication used for Oauth2 authentication
    """

    def __init__(self, authentication=None):
        self.authentication = authentication

    ##################
    ### Properties ###
    ##################
    @property
    def logger(self):
        """ the logging.Logger used for logging.
        Defaults to logging.getLogger(__name__).
        """
        try:  # try to return the internal property
            return self._logger
        except AttributeError:  # didn't work
            return logging.getLogger(__name__)  # return default logger

    @logger.setter
    def logger(self, logger):  # pragma: no cover
        assert isinstance(logger, logging.Logger), \
            "logger property has to be a logging.Logger"
        self._logger = logger

    @property
    def authentication(self):  # pragma: no cover
        """ The Authentication used for server authentication
        Defaults to an empty instance of Authentication on first use.
        """
        try:  # try to return the internal attribute
            return self._authentication
        except AttributeError:  # didn't work
            # set it to new instance
            self._authentication = authentication.Authentication()
        # return internal attribute
        return self._authentication

    @authentication.setter
    def authentication(self, newauth):
        if not isinstance(newauth, authentication.Authentication):
            self.logger.debug(
                "authentication property needs to be of "
                "class Authentication. Using empty instance instead "
                "of {}.".format(newauth))

            self._authentication = authentication.Authentication()
        else:
            self._authentication = newauth

    ###############
    ### Methods ###
    ###############
    def Getpublicdata(self, region, required_data=None, filter=False,
                      full=False, resolution=0.06):
        """ Issue a Getpublicdata POST request to the netatmo server

        Args:
            region (dict): dict definig the desired request region. Required
                keys: lat_ne [-85;85], lat_sw [-85;85], lon_ne [-180;180] and
                lon_sw [-180;180] with lat_ne > lat_sw and lon_ne > lon_sw
            required_data (str or None, optional): Defaults to None which means
                no filter.
            filter (bool, optional): server-side filter for stations with
                unusual data. Defaults to False with means no filter
            full (bool, optional): subdivide the requested region into smaller
                requests to get around the zoom levels present in the
                Getpublicdata API. This will take longer, obviously.
            resolution (float, optional): if full=True, specifies the maximum
                size of the lat/lon grid boxes used to subdivide the region in
                degrees. It seems like something like 0.06 degrees is the larges
                possible value to really get ALL stations. This is default.

        Returns:
            instance of GetpublicdataResponse with response data

        Raises:
            InvalidApiInputError or derivates if wrong input was provided
            ApiResponseError or derivates if api responded with error

        Returns:
            GetpublicdataResponse: The api response
        """

        ### Check the input ###
        # check the region
        for key, bounds in GETPUBLICDATA_REGION_BOUNDS.items():
            try:
                if region[key] < bounds[0] or region[key] > bounds[1]:
                    raise InvalidRegionError
            except BaseException:
                raise InvalidRegionError
        if region["lat_ne"] < region["lat_sw"] \
                or region["lon_ne"] < region["lon_sw"]:
            raise InvalidRegionError
        # check the required_data option
        if required_data not in GETPUBLICDATA_ALLOWED_REQUIRED_DATA \
                and required_data is not None:
            raise InvalidRequiredDataError
        # check the filter option
        if not isinstance(filter, bool):
            raise InvalidApiInputError("'filter' needs to be bool")

        if full:  # subdivide requests
            nlon = floor((region["lon_ne"] - region["lon_sw"]) / resolution)
            nlat = floor((region["lat_ne"] - region["lat_sw"]) / resolution)
            nlat = max(nlat, 2)
            nlon = max(nlon, 2)
            lats = np.linspace(region["lat_sw"], region["lat_ne"], num=nlat,)
            lons = np.linspace(region["lon_sw"], region["lon_ne"], num=nlon,)
            num_requests = (nlon - 1) * (nlat - 1)
            responses = []  # start with empty list
            request_count = 0
            for lat_sw, lat_ne, in zip(lats[:-1], lats[1:]):
                for lon_sw, lon_ne in zip(lons[:-1], lons[1:]):
                    region_cur = {"lat_sw": lat_sw, "lat_ne": lat_ne,
                                  "lon_sw": lon_sw, "lon_ne": lon_ne}
                    self.logger.info("current region: {}".format(region_cur))
                    self.logger.info("Request {} of {}...".format(
                        request_count + 1, num_requests))
                    response_cur = self.Getpublicdata(
                        region=region_cur, filter=filter, required_data=required_data, full=False)
                    self.logger.info("Stations in this region: {}".format(
                        response_cur.dataframe(only_inside=True).shape[0]))
                    responses.append(response_cur)
                    request_count += 1

            return responsetypes.GetpublicdataMultiResponse(
                request=requests.GetpublicdataRequest(),  # Fake request
                response={"parts": responses},
            )
        else:
            ### Create the payload ###
            payload = {}  # start with empty dict
            # set the access token
            payload["access_token"] = self.authentication.tokens.get(
                "access_token")
            # set the region
            payload.update(region)
            # set the required_data
            if required_data is not None:
                payload["required_data"] = str(required_data)
            # set the filter option
            payload["filter"] = str(filter).lower()

            ### issue the request ###
            apirequest = requests.GetpublicdataRequest(payload=payload)
            apiresponse = apirequest.response

            ### check for errors ###
            error = apiresponse.response.get("error", {}).get("message")
            if error:
                raise API_ERRORS.get(error, ApiResponseError(error))

            return apiresponse

    def Getmeasure(self, device_id, module_id=None, type=None,
                   scale=None, date_begin=None, date_end=None, real_time=False,
                   optimize=False, full=False):
        """ Issue a Getmeasure POST request to the netatmo server
        Taken from API documentation:
        https://dev.netatmo.com/dev/resources/technical/reference/common/getmeasure

        Args:
            device_id (str): The mac address of the device
            module_id (str, optinal): The mac address of the module of
                interest. If not specified, returns data of the device. If
                specified, returns data from the specified module
            type (list of str, optional): Measures interested in. List of
                "rain","humidity","pressure","wind" and "temperature".
            scale (str, optional): Timelapse between two measurements. "max"
                (every value is returned), "30min" (1 value every 30min),
                "1hour", "3hours", "1day", "1week", "1month". Defaults to "max".
            date_begin,date_end (int, optional): UNIX-timestamp of first/last
                measure to receive. Limit is 1024 measures.
            optimize (bool, optional): Determines the format of the answer.
                Default is False. For mobile apps we recommend True and False if
                bandwidth isn't an issue as it is easier to parse.
            real_time (bool, optional): If scale different than max, timestamps
                are by default offset + scale/2. To get exact timestamps, use
                true. Default is false.
            full (bool, optional): Make sure the full requested time range is
                returned by requesting multiple times. Default is ``False``.

        Returns:
            GetmeasureResponse: The api response
        """
        ### Check the input ###
        # check device_id
        if not MAC_ADDRESS_REGEX.match(device_id):
            raise InvalidApiInputError("device_id = '{}' doesn't look like "
                                       "MAC address".format(device_id))
        # check module_id
        if module_id is not None:
            if not MAC_ADDRESS_REGEX.match(module_id):
                raise InvalidApiInputError(
                    "module_id = '{}' doesn't look like "
                    "MAC address".format(module_id))
        # check scale
        if scale is not None:  # specified
            if scale not in GETMEASURE_ALLOWED_TYPES_BY_SCALE.keys():
                raise InvalidApiInputError("scale must be one of {}".format(
                    list(GETMEASURE_ALLOWED_TYPES_BY_SCALE.keys())))
        else:  # not specified
            scale = "max"  # use all values
        # check type
        allowed_types = GETMEASURE_ALLOWED_TYPES_BY_SCALE.get(scale, [])
        if type is not None:  # specified
            if not all([x in allowed_types for x in type]):
                raise InvalidApiInputError(
                    "at given scale '{}' type must be a "
                    "sublist of {}".format(
                        scale, allowed_types))
        else:  # not specified
            type = allowed_types  # use all
        # check date_begin
        if date_begin is not None:
            try:
                date_begin = int(date_begin)
                if date_begin < 0:
                    raise Exception
                if date_begin > time.time():
                    self.logger.warning(
                        "Specified date_begin is in the future!")
            except BaseException:
                raise InvalidApiInputError("date_begin must be UNIX-timestamp")
        if date_end is not None:
            try:
                date_end = int(date_end)
                if date_end < 0:
                    raise Exception
            except BaseException:
                raise InvalidApiInputError("date_end must be UNIX-timestamp")
        try:
            if date_end < date_begin:  # compare begin and end
                raise InvalidApiInputError("date_end must be greater than "
                                           "date_begin")
        except TypeError:
            pass
        # check the optimize option
        if not isinstance(optimize, bool):
            raise InvalidApiInputError("optimize must be bool")
        # check real_time option
        if not isinstance(real_time, bool):
            raise InvalidApiInputError("real_time must be bool")

        ### Prepare the payload ###
        payload = {}  # start with empty dict
        # access token
        payload["access_token"] = self.authentication.tokens.get(
            "access_token")
        # device id
        payload["device_id"] = device_id
        # module id
        if module_id is not None:
            payload["module_id"] = module_id
        # scale
        payload["scale"] = scale
        # type
        payload["type"] = ",".join(type)
        # date_begin
        if date_begin is not None:
            payload["date_begin"] = date_begin
        # date_end
        if date_end is not None:
            payload["date_end"] = date_end
        # optimize
        payload["optimize"] = str(optimize).lower()
        # real_time
        payload["real_time"] = str(real_time).lower()

        ### issue the request ###
        if full:
            responses = []
            last_max_time = payload["date_begin"]
            full_time_range = payload["date_end"] - payload["date_begin"]
            n_requests = 0
            while True:
                cur_payload = payload.copy()
                cur_payload.update({"date_begin":last_max_time})
                self.logger.info("Issuing Getmeasure request Nr. {}...".format(
                    n_requests+1))
                n_requests += 1
                apirequest = requests.GetmeasureRequest(payload=cur_payload)
                apiresponse = apirequest.response
                cur_times = \
                    [float(t) for t in apiresponse.response["body"].keys()]
                cur_max_time = max(cur_times)
                if cur_max_time >= payload["date_end"] \
                        or cur_max_time == last_max_time:
                    self.logger.info("Request Nr. {} didn't introduce "
                        "anything new. Stop.".format(n_requests))
                    break
                else:
                    self.logger.info("Request Nr. {:3d} fills up to {:5.1f}% of "
                        "the requested time range".format(n_requests,
                        (cur_max_time-payload["date_begin"]) /
                        full_time_range*100))
                    responses.append(apiresponse)
                last_max_time = cur_max_time
            return responsetypes.GetmeasureMultiResponse(
                request=requests.GetmeasureRequest(),  # Fake request
                response={"parts": responses},
            )
        else:
            apirequest = requests.GetmeasureRequest(payload=payload)
            apiresponse = apirequest.response

            ### check for errors ###
            error = apiresponse.response.get("error", {}).get("message")
            if error:
                raise API_ERRORS.get(error, ApiResponseError(error))

            return apiresponse

    def Getstationsdata(self, device_id, get_favourites=False):
        """ Issue a Getstationsdata POST request to the netatmo server
        Taken from API documentation:

        Args:
            device_id (str): The mac address of the device
            get_favourites (bool): To retrieve user's favorite weather stations.
                Is converted to bool. Default is false.

        Returns:
            GetstationsdataResponse: The api response
        """
        ### Check the input ###
        # check device_id
        if not MAC_ADDRESS_REGEX.match(device_id):
            raise InvalidApiInputError("device_id = '{}' doesn't look like "
                                       "MAC address".format(device_id))
        # check get_favourites
        try:
            get_favourites = bool(get_favourites)  # try to convert to bool
        except BaseException:
            raise InvalidApiInputError("get_favourites has to be bool or at "
                                       "least convertible to bool")

        ### create the payload ###
        payload = {}  # start with empty dict
        # access token
        payload["access_token"] = self.authentication.tokens.get(
            "access_token")
        # device id
        payload["device_id"] = device_id
        # get_favourites
        payload["get_favourites"] = str(get_favourites).lower()

        ### issue the request ###
        apirequest = requests.GetstationsdataRequest(payload=payload)
        apiresponse = apirequest.response

        ### check for errors ###
        error = apiresponse.response.get("error", {}).get("message")
        if error:
            raise API_ERRORS.get(error, ApiResponseError(error))

        return apiresponse

    def __repr__(self):  # pragma: no cover
        """ python representation of this object
        """
        # self.logger.debug("__repr__ called")
        reprstring = (
            "{classname}(\n"
            "authentication = {authentication}\n"
            ")").format(
            classname="{module}.{name}".format(
                name=self.__class__.__name__,
                module=self.__class__.__module__),
            authentication=self.authentication.__repr__())
        return reprstring



#!/usr/bin/env python3
# system modules
import logging

# external modules
import pandas as pd
from pandas import DataFrame
import numpy as np

# internal modules
from patatmo.api.errors import *


class ApiResponse(object):
    """ Base class for Netatmo api response datasets
    """

    def __init__(self, request, response):
        """ Class constructor
        Args:
            request (instance of ApiRequest or derivate): the api request
            response (dict): The raw api response
        """
        self.response = response
        self.request = request

    ##################
    ### Properties ###
    ##################
    @property
    def logger(self):  # pragma: no cover
        """ the logging.Logger used for logging.
        Defaults to logging.getLogger(__name__).
        """
        try:  # try to return the internal property
            return self._logger
        except AttributeError:  # didn't work
            return logging.getLogger(__name__)  # return default logger

    @logger.setter
    def logger(self, logger):  # pragma: no cover
        assert isinstance(logger, logging.Logger), \
            "logger property has to be a logging.Logger"
        self._logger = logger

    @property
    def response(self):
        """ The raw api response dict.
        """
        try:  # try to return internal attribute
            return self._response
        except AttributeError:  # pragma: no cover
            self._response = {}
        return self._response  # return internal attribute

    @response.setter
    def response(self, newresponse):
        assert isinstance(newresponse, dict), \
            "reponse property has to be of class dict"
        self._response = newresponse

    @property
    def request(self):
        """ The raw api request dict.
        """
        return self._request  # return internal attribute

    @request.setter
    def request(self, newrequest):
        self._request = newrequest

    def __repr__(self):  # pragma: no cover
        """ python representation of this object
        """
        # self.logger.debug("__repr__ called")
        reprstring = ("{classname}(\n"
                      "response = {response},\n"
                      "request = {request},\n"
                      ")").format(
            classname="{module}.{name}".format(
                name=self.__class__.__name__, module=self.__class__.__module__),
            # compact version
            response=self.response.__repr__(),
            # pretty version
            # response = json.dumps(self.response,sort_keys=True,indent=4)
            # the request
            request=self.request.__repr__()
        )
        return reprstring


class TokenResponse(ApiResponse):
    """ Class that holds the responded data of a Oauth2 token request
    """
    pass


class GetpublicdataResponse(ApiResponse):
    """ Class that holds the responded data of a Getstationdata request
    """

    def dataframe(self, only_inside=False):
        """
        Convert the response to a pandas.DataFrame

        Args:
            only_inside (bool, optional): Drop stations outside the requested
                region? Defaults to ``False``.

        Returns:
            pandas.DataFrame: The response converted to a DataFrame
        """
        # get the list of stations
        stations = self.response.get("body")
        if not isinstance(stations, list):  # pragma: no cover
            raise ApiResponseError("'body' part of response does not "
                                   "exist or is no list.")

        # start with basic dict
        stationdict = {
            "id": [],
            "altitude": [],
            "longitude": [],
            "latitude": [],
            "timezone": [],
        }
        # loop over all stations
        for station in stations:
            station_df = DataFrame()  # start with empty DataFrame

            ### gather general information ###
            station_id = station.get("_id", np.nan)
            # add id to dict
            stationdict["id"].append(station_id)

            ### gather positional information ###
            place = station.get("place", {})
            location = place.get("location", [np.nan, np.nan])
            longitude, latitude = location  # get longitude and latitude
            altitude = place.get("altitude", np.nan)
            timezone = place.get("timezone", np.nan)
            # add position to dict
            stationdict["altitude"].append(altitude)
            stationdict["longitude"].append(longitude)
            stationdict["latitude"].append(latitude)
            stationdict["timezone"].append(timezone)

            ### gather measurement information ###
            # loop over all measurements
            measdict = {}
            for module_id, measure in station.get("measures", {}).items():
                types = measure.get("type", [])
                res = measure.get("res", {})  # the time and values
                if not len(res) == 1:  # something is wrong
                    # self.logger.warning("module '{}' has not exactly one time! "
                    #     "Leaving it out.".format(module_id))
                    res = {np.nan: [np.nan] * len(types)}
                timestamp = list(res.keys())[0]
                measurements = res.get(timestamp)
                measdict.update(zip(types, measurements))
                timedict = {"time_{}".format(t): int(timestamp) for t in types}
                measdict.update(timedict)
                # self.logger.debug("measdict after update: {}".format(measdict))

            # add measurements to stationdict
            for key, val in measdict.items():
                try:
                    stationdict[key].append(val)
                except BaseException:
                    stationdict[key] = [val]

            # fill the remaining values
            length = max([len(x) for x in stationdict.values()])
            for key, val in stationdict.items():
                while len(val) < length:
                    val.append(np.nan)

        # create DataFrame
        df = DataFrame(stationdict)
        # convert times to datetime
        for col in df.columns:
            if col.startswith("time_"):
                df[col] = pd.to_datetime(df[col], unit="s", utc=True)

        # drop outliers if desired
        if only_inside:
            lat_ne = self.request.payload["lat_ne"]
            lon_ne = self.request.payload["lon_ne"]
            lat_sw = self.request.payload["lat_sw"]
            lon_sw = self.request.payload["lon_sw"]
            # TODO: What happens in other regions? Is outside still correct?
            outside = np.logical_or(
                df["latitude"] > lat_ne,
                np.logical_or(
                    df["latitude"] < lat_sw,
                    np.logical_or(
                        df["longitude"] > lon_ne,
                        df["longitude"] < lon_sw)))
            # drop outliers
            df.drop(np.where(outside)[0], inplace=True)

        # set index to device id
        if df.id.size != df.id.unique().size:
            self.logger.warning("Duplicate device IDs in the response!")
        df.set_index(df.id, inplace=True)
        # return the resulting DataFrame
        return df


class GetpublicdataMultiResponse(GetpublicdataResponse):
    """ Class that holds responded data of a subdivided Getpublicdata request
    """

    def dataframe(self, only_inside=False):
        """
        Convert the multiple subdivided Getpublicdata responses to a single
        dataframe

        Returns:
            pandas.Dataframe: the merged dataframe
        """
        try:
            count_stations = 0
            responses = self.response.get("parts")
            for response in responses:
                df_cur = response.dataframe(only_inside=only_inside)
                count_stations += df_cur.shape[0]
                try:
                    # join datasets
                    df = pd.concat([df, df_cur], axis=0, join="outer")
                    # df = df.combine_first(df_cur)
                    # drop duplicates
                except NameError:  # first one
                    df = df_cur
            # drop duplicates
            df.drop_duplicates(subset="id", inplace=True)
            return df  # return
        except TypeError:  # not iterable e.g.
            raise ApiResponseError(
                "subdivided Getpublicdata response contains bogus")


class GetmeasureResponse(ApiResponse):
    """ Class that holds the responded data of a Getmeasure request
    """

    def dataframe(self):
        """
        Convert the response to a pandas.DataFrame

        Returns:
            pandas.DataFrame: The response converted to a DataFrame
        """
        body = self.response.get("body")
        if isinstance(body, list):  # optimized # pragma: no cover
            if body:  # only if there really is something
                raise NotImplementedError("converting 'optimized' Getmeasure "
                                          "response is not yet implemented")
            else:
                body = {}  # fake to empty dict
        if isinstance(body, dict):  # unoptimized
            try:
                types = self.request.payload.get(
                    "type").split(",")  # the types
            except AttributeError:  # pragma: no cover
                raise InvalidApiInputError(
                    "There is no sensible 'type' "
                    "section in the request's payload. Strange...")

            # start with empty measurement dict
            measdict = {"time": []}
            # loop over all time-measurement pairs
            for timestamp, measurements in body.items():
                if not len(types) == len(measurements):  # pragma: no cover
                    raise ApiResponseError(
                        "number of requested types does not "
                        "match number of responded types")
                d = dict(zip(types, measurements))
                d.update({"time": int(timestamp)})
                # add measurements to measdict
                for key, val in d.items():
                    try:
                        measdict[key].append(val)
                    except BaseException:
                        measdict[key] = [val]

        else:  # bullshit # pragma: no cover
            raise ApiResponseError("'body' part of response does not "
                                   "exist or is neither list not dict.")

        # create DataFrame
        df = DataFrame(measdict)
        # convert times to datetime
        for col in df.columns:
            if col.startswith("time"):
                df[col] = pd.to_datetime(df[col], unit="s", utc=True)
        # sort the data frame by time
        df = df.sort_values(by="time")
        # index
        df.set_index("time", inplace=True)
        # return the resulting DataFrame
        return df


class GetmeasureMultiResponse(GetmeasureResponse):
    """
    Class holding responded data of multiple Getmeasure requests
    """
    def dataframe(self):
        try:
            responses = self.response.get("parts")
            for response in responses:
                df_cur = response.dataframe()
                try:
                    # join datasets
                    df = pd.concat([df, df_cur], axis=0, join="outer")
                except NameError:  # first one
                    df = df_cur
            # drop duplicates
            df.drop_duplicates(inplace=True)
            # sort
            df.sort_index(inplace=True)
            return df  # return
        except TypeError:  # not iterable e.g.
            raise ApiResponseError(
                "subdivided Getmeasure response contains bogus")

class GetstationsdataResponse(ApiResponse):
    """ Class that holds the responded data of a Getstationsdata request
    """

    def dataframe(self):
        """
        Convert the response to a pandas.DataFrame

        Returns:
            pandas.DataFrame: The response converted to a DataFrame
        """
        body = self.response.get("body")
        if not isinstance(body, dict):  # optimized # pragma: no cover
            raise ApiResponseError("api response body is no dict. Strange...")
        devices = body.get("devices")
        if not isinstance(devices, list):  # no list # pragma: no cover
            raise ApiResponseError("api response devices part is no list. "
                                   "Strange...")
        if len(devices) > 1:  # pragma: no cover
            raise ApiResponseError("api response devices list is longer than "
                                   "one. Strange...")
        try:
            dashboard_data = devices[0]["dashboard_data"]  # device
        except KeyError:  # pragma: no cover
            raise ApiResponseError("api response devices list entry has no "
                                   "dashboard_data")

        # get proper data
        data = {}  # start with empty dict
        for key, val in dashboard_data.items():
            if isinstance(
                    val, dict) or isinstance(
                    val, list):  # pragma: no cover
                raise ApiResponseError("dashboard_data values are not scalars")
            data[key] = [val]  # put it into a list

        # convert to data frame
        df = pd.DataFrame(data)

        # reset index
        df.reset_index(inplace=True)

        return df



# -*- coding: utf-8 -*-
# System modules

# External modules

# Internal modules

import patatmo.api.authentication
import patatmo.api.client
import patatmo.api.errors
import patatmo.api.variables
import patatmo.api.requests
import patatmo.api.responsetypes

__all__ = ['authentication']



#!/usr/bin/env python3
# module for api errors

##############
### Errors ###
##############


class ApiResponseError(BaseException):
    pass


class InvalidCredentialsError(ApiResponseError):
    pass


class InvalidApiInputError(BaseException):
    pass


class InvalidRegionError(InvalidApiInputError):
    def __init__(self):
        message = \
            ("'region' required keys: "
             "lat_ne [-85;85], lat_sw [-85;85], "
             "lon_ne [-180;180] and lon_sw [-180;180] "
             "with lat_ne > lat_sw and lon_ne > lon_sw")
        super().__init__(message)


class InvalidRequiredDataError(InvalidApiInputError):
    def __init__(self):
        message = "'required_data' must be None or in {}".format(
            GETPUBLICDATA_ALLOWED_REQUIRED_DATA)
        super().__init__(message)


class InvalidApiRequestInputError(BaseException):
    pass


class InvalidPayloadError(InvalidApiRequestInputError):
    pass


API_ERRORS = {
    "invalid_client": InvalidCredentialsError("wrong credentials"),
    "invalid_request": ApiResponseError("invalid request"),
    "invalid_grant": ApiResponseError("invalid grant"),
    "Device not found": ApiResponseError("Device not found - Check Device ID "
                                         "and permissions")
}



# -*- coding: utf-8 -*-
# System modules
import os
import sys
import argparse
import logging
import configparser

# External modules

# Internal modules
from ..api.client import NetatmoClient
from ..api.authentication import Authentication


def cli_argumentparser(*args, **kwargs):  # pragma: no cover
    """
    Set up and return an :any:`argparse.ArgumentParser` with default arguments
    every patatmo cli script needs.

    Args:
        args (list, optional): further positional arguments handed to the
            constructor
        kwargs (dict, optional): further keyword arguments handed to the
            constructor

    Returns:
        argparse.ArgumentParser: the parser with default arguments
    """
    parser = argparse.ArgumentParser(*args, **kwargs)
    parser.add_argument(
        "-p",
        "--password",
        help="Netatmo developer account password. "
        "Takes precedence over NETATMO_PASSWORD environment variable",
        required=False)
    parser.add_argument(
        "-u",
        "--user",
        help="Netatmo developer account username. "
        "Takes precedence over NETATMO_USERNAME environment variable",
        required=False)
    parser.add_argument(
        "-i",
        "--id",
        help="Netatmo app client id. "
        "Takes precedence over NETATMO_CLIENT_ID environment variable",
        required=False)
    parser.add_argument(
        "-s",
        "--secret",
        help="Netatmo app client secret. "
        "Takes precedence over NETATMO_CLIENT_SECRET environment variable",
        required=False)
    parser.add_argument(
        "-o",
        "--output",
        help="output file. Defaults to '-' which means STDOUT.",
        default="-")
    parser.add_argument(
        "-f",
        "--tmpfile",
        help="temporary authentication file",
        required=False)
    parser.add_argument(
        "--noconfig",
        help="ignore the user's personal "
        "patatmo configuration folder completely",
        required=False,
        default=False,
        action="store_true")
    parser.add_argument("-v", "--verbose", help="verbose output",
                        action="store_true", default=False)
    parser.add_argument("--debug", help="even more verbose output",
                        action="store_true", default=False)
    return parser


def get_credentials(arguments=argparse.Namespace()):  # pragma: no cover
    """
    Try to determine the netatmo credentials in this order:

    1) command-line arguments
    2) environment variables
    3) user configuration folder

    Raise an error if it was not possible to determine all credentials.

    Args:
        arguments (argparse.Namespace, optional): the parsed command-line
            arguments.

    Returns:
        dict : the credentials ready to use by :any:`Authentication`.

    Raises:
        CredentialsError : if any of the credentials could not be determined
    """
    def get_credential(longname, argname, confname, envname):
        try:  # try to get
            value = getattr(arguments, argname)
            return(value.strip())
        except AttributeError:  # user not given via command-line
            value = os.environ.get(envname)
            try:  # check if value is a string
                return(value.strip())
            except AttributeError:  # not given in environment
                try:  # try to read from config file
                    config = get_configuration()  # get user configuration
                    return(config["account"].get(confname).strip())
                except (KeyError, AttributeError):  # not in configuration file
                    methods = []
                    if not arguments == argparse.Namespace():
                        methods.append("command-line argument '{}'".format(
                            argname))
                    methods.append("environment variable '{}'".format(envname))
                    if not (arguments.noconfig
                            if hasattr(arguments, "noconfig") else False):
                        methods.append(
                            "'{}' key in 'accounts' section "
                            "in patatmo user configuration file '{}'".format(
                                confname, get_configuration_file()))
                    if not methods:
                        methods.append("nothing :-) You should not see this.")

                    methodstr = " or ".join(
                        x for x in [", ".join(methods[0:-1]), methods[-1]] if x)
                    message = "Specify {} via {}".format(longname, methodstr)
                    raise CredentialsError(message)

    needed_credentials = [
        ["Netatmo account username", "user", "username", "NETATMO_USERNAME",
            "username"],
        ["Netatmo account password", "password", "password", "NETATMO_PASSWORD",
            "password"],
        ["Netatmo app id", "id", "client_id", "NETATMO_CLIENT_ID", "client_id"],
        ["Netatmo app secret", "secret", "client_secret", "NETATMO_CLIENT_SECRET",
            "client_secret"],
    ]

    credentials = {}
    for longname, argname, confname, envname, resname in needed_credentials:
        credentials[resname] = get_credential(
            longname=longname,
            argname=argname,
            envname=envname,
            confname=confname)

    return credentials


def get_client(arguments=argparse.Namespace()):  # pragma: no cover
    """
    Set up a client for direct usage.

    Args:
        arguments (argparse.Namespace, optional): the parsed command-line
            arguments.

    Returns:
        NetatmoClient : a ready-to-use netatmo api client

    Raises:
        CredentialsError : if any of the credentials could not be determined
    """
    # read credentials
    try:
        credentials = get_credentials(arguments=arguments)
    except CredentialsError as e:
        logging.error(str(e))
        sys.exit(1)

    if not (arguments.noconfig if hasattr(arguments, "noconfig") else False):
        configdir = get_configuration_dir()
        make_sure_dir_exists(configdir)
        tmpfilename = \
            ".{}.json".format(credentials.get("client_id", "auth").strip())
        tmpfile = os.path.join(configdir, tmpfilename)
    else:
        tmpfile = (
            arguments.tmpfile if hasattr(
                arguments,
                "tmpfile") else None)

    # set up a client
    client = NetatmoClient(
        authentication=Authentication(
            credentials=credentials,
            tmpfile=tmpfile,
        ),
    )

    return client


def get_output_file(arguments=argparse.Namespace()):  # pragma: no cover
    """
    Return a file handler according to the 'output' option in the given
    arguments.

    Args:
        arguments (argparse.Namespace, optional): the parsed command-line
            arguments.

    Returns:
        filehandle : the filehandle to output to
    """
    try:
        output = arguments.output
    except AttributeError:
        output = "-"
    return sys.stdout if output == '-' or not output else open(output, "w")


def get_configuration_dir():  # pragma: no cover
    """
    Return the user's configuration directory

    Returns:
        str : path to the user's configueration directory
    """
    return os.path.expanduser("~/.patatmo")


def make_sure_dir_exists(directory):  # pragma: no cover
    """
    Make sure the given directory exists

    Args:
        directory (str): the path of interest
    """
    if not os.path.exists(directory):
        os.makedirs(directory)


def get_configuration_file():  # pragma: no cover
    """
    Return the user's configuration file

    Returns:
        str : path to the user's configueration file
    """
    return os.path.join(get_configuration_dir(), "settings.conf")


def get_configuration():  # pragma: no cover
    """
    Return the user's patatmo configuration

    Returns:
        configparser.ConfigParser: the user's patatmo configuration
    """
    config = configparser.ConfigParser()
    config.read(get_configuration_file())
    return config


class CredentialsError(ValueError):  # pragma: no cover
    pass
